#ifndef hilbert_h 	/* include this only once */
#define hilbert_h 1

/* definition of the base and the operators of the Hilbert space */

#include "modeldef.h"
#include "defs.h"
#include "spins.h"

/* definition of the vakuum state */
#define TBASE0 ((tbase)0)

/* ================= inline - functions ======================= */

/* this function is only for better readability */
/* function returns up and down spin of site i, 0=empty, 1=up, ... */
static inline int geti(tbase cfg,int i){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return 1+(1&(int)(cfg>>i));}
#else
  return   (3&(int)(cfg>>(2*i)));}
#endif

/* function set up and/or down spin on site i, 0=empty, 1=up, ... */
static inline tbase puti(tbase cfg,int i,int s){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return (cfg & ~((tbase)1<<i)) | ((tbase)(s-1))<<i; }
#else
  return (cfg & ~(((tbase)3)<<(2*i))) | ((tbase)s)<<(2*i);}
#endif
        

/* fast encode ud[] to cfg, 0 1=u 2=d 3=du ud[0]=lowest Dez01 */
static inline void encode_ud(tbase *cfg, t_sites ud[NN], int nn){
  int i; *cfg=TBASE0;
#if CM(1,0,0)
  for(i=nn-1;i>=0;i--) (*cfg) = ((*cfg)<<1) | ((tbase)(ud[i]-1));
#else
  for(i=0;i<nn;i++)   (*cfg) |= ((tbase)ud[i])<<(2*i);
#endif
}

#if CM(0,1,1)
/* fast xor single spin i ud=0: u, ud=1: d, Apr02 */
static inline tbase tbase_xor(tbase cfg, int i, int ud){
#if CM(1,0,0)
  cfg ^= (((tbase)1)<<i);    /* ud not used!? */
#else
  cfg ^= (((tbase)1)<<(2*i+ud));
#endif
  return cfg;
}
#endif

int b_out(struct fullsbase *b,int len);
void   paini();
ULong numsymconf();
/* const ULong numconf(const int nn, const int nu, const int nd); */
void speed_tests(void *ptr);
void ini_norm2(void);
tbase  ll1(ULong ii); /* only for ll1(0) ??? */
int initmasks(int nn);

ULong scfg2idx(tbase i);  /* (base_config=scfg)-to-index, 0 <= idx < node_n1 */
int   scfg2node(tbase i);/* get global mpi node number from sym-config */
char *ccfg(tbase ic); /* Spinconfiguration |..43210> zB: |...o.xoQx> */


static inline  /* speedup for expectation + tU-sign 2016-01 */
int zi(tbase cfg,int i){   /* Siz -> -1,0,1, moved from spins.c 2016-01 */
#if CM(1,0,0)
  return 1-2*(1&(int)(cfg>>i));}
#else
  int i1=geti(cfg,i); return( (i1==2) ? -1 : (i1==1) ? 1 : 0);}
#endif

static inline
int zizj(tbase cfg,int i,int j){   /* Siz*Sjz -> -1,0,1 */
#if CM(1,0,0)
  return 1-2*(1&(((int)(cfg>>i))^(int)(cfg>>j)));}
#else
  int i1=geti(cfg,i)*geti(cfg,j); return((i1==2)?-1:(i1==4 || i1==1)?1:0);}
#endif


/* 2015-02:  b_getbase renamed to allsymcfg */
tbase allsymcfg(const tbase ib1,struct fullsbase *b);

/* --- ToDo: define fastest variant of minsymcfg_* as default here: --- */
/*  this is very CPU and compiler dependent (SIMD, pipelines etc.)
best-choice table, time in CPU-clks per symmetry, smaller is better 2015-06
CPU \ ALGO        lNv1  lNv8   bn1     bn8 (clks/permutation)
  1ALU^SHFT        160   160   110     81.1 clks 11*(7+3/8)   IPC=1
  1ALU+SHFT         82    80    77     46.8 clks              IPC=1.7
  1ALU^SHFT+1ALU    82    80    55     40.6 clks              IPC=2
  2(ALU^SHFT)       82    80    55     40.6 clks = 1*SSE2-ALU IPC=2
  2ALUs+SHIFT       82          57     39?                    IPC=2.1
  3ALUs             43    54?   40     27.0 clks              IPC=3
  3ALU128           22          20     13.5 clks              IPC=6
  3ALU256           11          10      6.8 clks              IPC=12 (i64)
  1SHFT256+2LOG256+ 21    16?   11      8.8 clks +1shft^alu   IPC=9.2 (i64)

2015-10-08 + factor 2 correction (ToDo: add lNvs? sun?)
CPU \ ALGO   lNvs lNb14 19 VS bnVS     bestChoice(benchmarks) IPC=81i/bnVclks
  MIPS-r4k   192   43  47  v4  93   clks lNbrk SiCortex 5Kf IPC=0.9
  AtomN455   141   44  51 v16  44   clks bn16  SSE*8pipe    IPC=1.8 in-order
    pCore=HT 128   32  35  v8  30   clks bn8   2 threads, per core, IPC=2.7
  k8          87   24  30  v8  41   clks lNbrk comp1-2.6GHz=k10exe  IPC=1.9
  k10              24  28  v8  26   clks bn8   comp2-2.3GHz,meggie  IPC=3.1
  iCore2      60   24  30  v8  23.5 clks bn8   JS-PC-2.7GHz_@1.6GHz IPC=3.4
  sandybrdg.       20  26  v8  20   clks bn8   E54620-2.2GHz        IPC=4.05
  haswell     36   16  20 v32  10.5 clks bn32  AVX2*8pipe           IPC=7.7=max2pipes*4simd*64bit v3=16FLOP64/clk
  AVX512+icc       62  70 v16  54   clks bn64? 1thread! XeonPhi IPC=3? bad?
     pC=4HTs  ?                 ?            57*4HT XeonPhi in-order ToDo asm?
  # haswell-advantage=2.4x using vectorized benes-algo, bnVS/lNvs=ca3.5
  # speedtest used, the SH-loop is usually 2-3fold slower (why?)
  # lNbrk: lNb14=14+26 lNb19=19+21
  # itanium 4DP/clk (usually linpack means DP O(n^3), but SP version exist)
  # 3 linpacks: n=100 n=1000 +highly-parallel(HPL)=top500=64bit
 */
#if 1 && NN<=CM(512,0,0) && !defined(__R4000__) // 2015-07-06 NN<=64, ToDo: 512
// && !defined(__k8__)  // k8 is set on other platforms too p.e. sandybridge 
// ToDo: bug in NN>64 (NN=92 ud=6+86 lc.gpl)
#define CFG_USE_BENES 1  /* 0=old_lNbrk_code 1=new_bn_code vectorizable */
#else
#define CFG_USE_BENES 0  /* ToDo: fermionic sign for benes network */
#endif
/* ToDo2015: add tJ, tU, + JJ NN>64, 2015-09-23 test.JJ128 */
#if     CFG_USE_BENES  /* faster for old CPUs and small NN */
#define minsymcfg_dflt           minsymcfg_lNbrk  /* no bn available, unused */
#define minsymcfg_blk_dflt       minsymcfg_blk_bnVS
#define minsymcfg_blk_dflt_no_ud minsymcfg_blk_bnVS_no_ud
#else /* faster on vector CPUs or AVX2 (untested) and big NN (64++) */
#define minsymcfg_dflt           minsymcfg_lNbrk     /* may be slow */
#define minsymcfg_blk_dflt       minsymcfg_blk_lNbrk
#define minsymcfg_blk_dflt_no_ud minsymcfg_blk_lNbrk /* missing _no_ud */
#endif

/* icc16 warns "const" as meaningless, we do not really need it */ 
/* const */
tbase minsymcfg_dflt(const tbase ib1, struct sbase *b); /* single config */
void  minsymcfg_blk_dflt(thbuf *hbuf);        /* cfg block (vectorized) */
void  minsymcfg_blk_dflt_no_ud(thbuf *hbuf);

#if CM(0,1,1)
/*
 * this comes from fermionic sign (-1)^Sum(n(i<j)) for site j
 *   we calculate a product for 2 sites here
 * Is parity base coding dependend? => only mask depends from base code!
 * +1 if even number of 1-bits,  -1 else
 * inline -> code inserted (faster,only -O and C++)
 * static -> function not externaly available    
 */
static inline int parity(tbase1 ib1){ /* odd:-1=1111, even:+1=0001 */
#if NN>CM(64,32,32)
    int i;              /* this works only with class vlint */
    for(i=64;i<CM(nn,2*nn,2*nn);i<<=1) ib1^=ib1>>i; /* not testet ??? */
#endif
#if NN>CM(32,16,16)
    ib1^=ib1>>32;
#endif
    ib1^=ib1>>16; ib1^=ib1>>8; ib1^=ib1>>4;
    ib1^=ib1>>2;  ib1^=ib1<<1; return 1-(2&(int)ib1); /* 6 OPs */
/*  return 1-(((0x6996 >> int(ib1&0xf))<<1) & 2); */  /* 5 OPs */
/*  return 1-( (0xd32c >> int(ib1&0xf))     & 2); */  /* 4 OPs 32bit ToDo */
}
#endif /* CM */

/* see spins.c, depends from base coding (code2)
 *   signum between two sites, for one site use p_mask[j][0] ???
 */
extern tbase1 p_mask[NN][NN];

/* return a unique(?) list of counted different pairs = ising energy index
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode==0: output list, free list, return 0
   mode==1: return list
   buf = return space OR NULL for malloc
 */
int *get_dcorr(tbase b1, int mode, int *buf); /* Apr06, v2.25 */
    

#endif
